@@ -1,19 +1,10 @@
 // Copyright (C) 2020 Mathieu Duponchelle <mathieu@centricular.com>
 //
-// This library is free software; you can redistribute it and/or
-// modify it under the terms of the GNU Library General Public
-// License as published by the Free Software Foundation; either
-// version 2 of the License, or (at your option) any later version.
+// This Source Code Form is subject to the terms of the Mozilla Public License, v2.0.
+// If a copy of the MPL was not distributed with this file, You can obtain one at
+// <https://mozilla.org/MPL/2.0/>.
 //
-// This library is distributed in the hope that it will be useful,
-// but WITHOUT ANY WARRANTY; without even the implied warranty of
-// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-// Library General Public License for more details.
-//
-// You should have received a copy of the GNU Library General Public
-// License along with this library; if not, write to the
-// Free Software Foundation, Inc., 51 Franklin Street, Suite 500,
-// Boston, MA 02110-1335, USA.
+// SPDX-License-Identifier: MPL-2.0
 
 use gst::glib;
 use gst::prelude::*;
@@ -26,7 +17,7 @@ use gst::{
 use std::default::Default;
 
 use rusoto_core::Region;
-use rusoto_credential::{ChainProvider, ProvideAwsCredentials};
+use rusoto_credential::{ChainProvider, ProvideAwsCredentials, StaticProvider};
 
 use rusoto_signature::signature::SignedRequest;
 
@@ -127,6 +118,8 @@ struct Settings {
     vocabulary: Option<String>,
     session_id: Option<String>,
     results_stability: AwsTranscriberResultStability,
+    access_key: Option<String>,
+    secret_access_key: Option<String>,
 }
 
 impl Default for Settings {
@@ -138,6 +131,8 @@ impl Default for Settings {
             vocabulary: None,
             session_id: None,
             results_stability: DEFAULT_STABILITY,
+            access_key: None,
+            secret_access_key: None,
         }
     }
 }
@@ -375,6 +370,22 @@ impl Transcriber {
     ) {
         let lateness = self.settings.lock().unwrap().lateness;
 
+        if alternative.items.len() <= state.partial_index {
+            gst_error!(
+                CAT,
+                obj: element,
+                "sanity check failed, alternative length {} < partial_index {}",
+                alternative.items.len(),
+                state.partial_index
+            );
+
+            if !partial {
+                state.partial_index = 0;
+            }
+
+            return;
+        }
+
         for item in &alternative.items[state.partial_index..] {
             let start_time =
                 gst::ClockTime::from_nseconds((item.start_time as f64 * 1_000_000_000.0) as u64)
@@ -437,7 +448,7 @@ impl Transcriber {
                 );
 
                 let caps = gst::Caps::builder("text/x-raw")
-                    .field("format", &"utf8")
+                    .field("format", "utf8")
                     .build();
                 events.push(
                     gst::event::Caps::builder(&caps)
@@ -588,18 +599,15 @@ impl Transcriber {
                 }
             };
 
-            let transcribe = Self::from_instance(&element);
-            match transcribe.loop_fn(&element, &mut receiver) {
-                Err(err) => {
-                    element_error!(
-                        &element,
-                        gst::StreamError::Failed,
-                        ["Streaming failed: {}", err]
-                    );
-                    let _ = transcribe.srcpad.pause_task();
-                }
-                Ok(_) => (),
-            };
+            let transcribe = element.imp();
+            if let Err(err) = transcribe.loop_fn(&element, &mut receiver) {
+                element_error!(
+                    &element,
+                    gst::StreamError::Failed,
+                    ["Streaming failed: {}", err]
+                );
+                let _ = transcribe.srcpad.pause_task();
+            }
         });
         if res.is_err() {
             return Err(loggable_error!(CAT, "Failed to start pad task"));
@@ -759,9 +767,7 @@ impl Transcriber {
                 let running_time = state.in_segment.to_running_time(buffer.pts());
                 let now = element.current_running_time();
 
-                delay = running_time
-                    .zip(now)
-                    .and_then(|(running_time, now)| running_time.checked_sub(now));
+                delay = running_time.opt_checked_sub(now).ok().flatten();
             }
         }
 
@@ -860,15 +866,39 @@ impl Transcriber {
 
         gst_info!(CAT, obj: element, "Connecting ..");
 
-        let creds = {
-            let _enter = RUNTIME.enter();
-            futures::executor::block_on(ChainProvider::new().credentials()).map_err(|err| {
-                gst_error!(CAT, obj: element, "Failed to generate credentials: {}", err);
-                error_msg!(
-                    gst::CoreError::Failed,
-                    ["Failed to generate credentials: {}", err]
-                )
-            })?
+        let creds = match (
+            settings.access_key.as_ref(),
+            settings.secret_access_key.as_ref(),
+        ) {
+            (Some(access_key), Some(secret_access_key)) => {
+                let _enter = RUNTIME.enter();
+                futures::executor::block_on(
+                    StaticProvider::new_minimal(access_key.clone(), secret_access_key.clone())
+                        .credentials()
+                        .map_err(|err| {
+                            gst_error!(
+                                CAT,
+                                obj: element,
+                                "Failed to generate credentials: {}",
+                                err
+                            );
+                            error_msg!(
+                                gst::CoreError::Failed,
+                                ["Failed to generate credentials: {}", err]
+                            )
+                        }),
+                )?
+            }
+            _ => {
+                let _enter = RUNTIME.enter();
+                futures::executor::block_on(ChainProvider::new().credentials()).map_err(|err| {
+                    gst_error!(CAT, obj: element, "Failed to generate credentials: {}", err);
+                    error_msg!(
+                        gst::CoreError::Failed,
+                        ["Failed to generate credentials: {}", err]
+                    )
+                })?
+            }
         };
 
         let language_code = settings
@@ -933,7 +963,7 @@ impl Transcriber {
         let element_weak = element.downgrade();
         let future = async move {
             while let Some(element) = element_weak.upgrade() {
-                let transcribe = Self::from_instance(&element);
+                let transcribe = element.imp();
                 let msg = match ws_stream.next().await {
                     Some(msg) => msg,
                     None => {
@@ -1065,7 +1095,7 @@ impl ObjectImpl for Transcriber {
     fn properties() -> &'static [glib::ParamSpec] {
         static PROPERTIES: Lazy<Vec<glib::ParamSpec>> = Lazy::new(|| {
             vec![
-                glib::ParamSpec::new_string(
+                glib::ParamSpecString::new(
                     "language-code",
                     "Language Code",
                     "The Language of the Stream, see \
@@ -1074,7 +1104,7 @@ impl ObjectImpl for Transcriber {
                     Some("en-US"),
                     glib::ParamFlags::READWRITE | gst::PARAM_FLAG_MUTABLE_READY,
                 ),
-                glib::ParamSpec::new_uint(
+                glib::ParamSpecUInt::new(
                     "latency",
                     "Latency",
                     "Amount of milliseconds to allow AWS transcribe",
@@ -1083,7 +1113,7 @@ impl ObjectImpl for Transcriber {
                     DEFAULT_LATENCY.mseconds() as u32,
                     glib::ParamFlags::READWRITE | gst::PARAM_FLAG_MUTABLE_READY,
                 ),
-                glib::ParamSpec::new_uint(
+                glib::ParamSpecUInt::new(
                     "lateness",
                     "Lateness",
                     "Amount of milliseconds to introduce as lateness",
@@ -1092,7 +1122,7 @@ impl ObjectImpl for Transcriber {
                     DEFAULT_LATENESS.mseconds() as u32,
                     glib::ParamFlags::READWRITE | gst::PARAM_FLAG_MUTABLE_READY,
                 ),
-                glib::ParamSpec::new_string(
+                glib::ParamSpecString::new(
                     "vocabulary-name",
                     "Vocabulary Name",
                     "The name of a custom vocabulary, see \
@@ -1101,14 +1131,14 @@ impl ObjectImpl for Transcriber {
                     None,
                     glib::ParamFlags::READWRITE | gst::PARAM_FLAG_MUTABLE_READY,
                 ),
-                glib::ParamSpec::new_string(
+                glib::ParamSpecString::new(
                     "session-id",
                     "Session ID",
                     "The ID of the transcription session, must be length 36",
                     None,
                     glib::ParamFlags::READWRITE | gst::PARAM_FLAG_MUTABLE_READY,
                 ),
-                glib::ParamSpec::new_enum(
+                glib::ParamSpecEnum::new(
                     "results-stability",
                     "Results stability",
                     "Defines how fast results should stabilize",
@@ -1116,6 +1146,20 @@ impl ObjectImpl for Transcriber {
                     DEFAULT_STABILITY as i32,
                     glib::ParamFlags::READWRITE | gst::PARAM_FLAG_MUTABLE_READY,
                 ),
+                glib::ParamSpecString::new(
+                    "access-key",
+                    "Access Key",
+                    "AWS Access Key",
+                    None,
+                    glib::ParamFlags::READWRITE | gst::PARAM_FLAG_MUTABLE_READY,
+                ),
+                glib::ParamSpecString::new(
+                    "secret-access-key",
+                    "Secret Access Key",
+                    "AWS Secret Access Key",
+                    None,
+                    glib::ParamFlags::READWRITE | gst::PARAM_FLAG_MUTABLE_READY,
+                ),
             ]
         });
 
@@ -1168,6 +1212,14 @@ impl ObjectImpl for Transcriber {
                     .get::<AwsTranscriberResultStability>()
                     .expect("type checked upstream");
             }
+            "access-key" => {
+                let mut settings = self.settings.lock().unwrap();
+                settings.access_key = value.get().expect("type checked upstream");
+            }
+            "secret-access-key" => {
+                let mut settings = self.settings.lock().unwrap();
+                settings.secret_access_key = value.get().expect("type checked upstream");
+            }
             _ => unimplemented!(),
         }
     }
@@ -1198,11 +1250,21 @@ impl ObjectImpl for Transcriber {
                 let settings = self.settings.lock().unwrap();
                 settings.results_stability.to_value()
             }
+            "access-key" => {
+                let settings = self.settings.lock().unwrap();
+                settings.access_key.to_value()
+            }
+            "secret-access-key" => {
+                let settings = self.settings.lock().unwrap();
+                settings.secret_access_key.to_value()
+            }
             _ => unimplemented!(),
         }
     }
 }
 
+impl GstObjectImpl for Transcriber {}
+
 impl ElementImpl for Transcriber {
     fn metadata() -> Option<&'static gst::subclass::ElementMetadata> {
         static ELEMENT_METADATA: Lazy<gst::subclass::ElementMetadata> = Lazy::new(|| {
@@ -1220,7 +1282,7 @@ impl ElementImpl for Transcriber {
     fn pad_templates() -> &'static [gst::PadTemplate] {
         static PAD_TEMPLATES: Lazy<Vec<gst::PadTemplate>> = Lazy::new(|| {
             let src_caps = gst::Caps::builder("text/x-raw")
-                .field("format", &"utf8")
+                .field("format", "utf8")
                 .build();
             let src_pad_template = gst::PadTemplate::new(
                 "src",
@@ -1231,9 +1293,9 @@ impl ElementImpl for Transcriber {
             .unwrap();
 
             let sink_caps = gst::Caps::builder("audio/x-raw")
-                .field("format", &"S16LE")
-                .field("rate", &gst::IntRange::<i32>::new(8000, 48000))
-                .field("channels", &1)
+                .field("format", "S16LE")
+                .field("rate", gst::IntRange::new(8000i32, 48000))
+                .field("channels", 1)
                 .build();
             let sink_pad_template = gst::PadTemplate::new(
                 "sink",
