@@ -1,34 +1,38 @@
 // Copyright (C) 2019 Amazon.com, Inc. or its affiliates <mkolny@amazon.com>
 //
-// Licensed under the Apache License, Version 2.0 <LICENSE-APACHE or
-// http://www.apache.org/licenses/LICENSE-2.0> or the MIT license
-// <LICENSE-MIT or http://opensource.org/licenses/MIT>, at your
-// option. This file may not be copied, modified, or distributed
-// except according to those terms.
+// This Source Code Form is subject to the terms of the Mozilla Public License, v2.0.
+// If a copy of the MPL was not distributed with this file, You can obtain one at
+// <https://mozilla.org/MPL/2.0/>.
+//
+// SPDX-License-Identifier: MPL-2.0
 
 use gst::glib;
 use gst::prelude::*;
 use gst::subclass::prelude::*;
-use gst::{gst_debug, gst_error, gst_info, gst_trace};
+use gst::{gst_debug, gst_error, gst_info, gst_log, gst_trace, gst_warning};
 
 use gst_base::subclass::prelude::*;
 
 use futures::future;
-use rusoto_core::region::Region;
+use rusoto_core::{region::Region, request::HttpClient};
+use rusoto_credential::StaticProvider;
 use rusoto_s3::{
-    CompleteMultipartUploadRequest, CompletedMultipartUpload, CompletedPart,
-    CreateMultipartUploadRequest, S3Client, UploadPartRequest, S3,
+    AbortMultipartUploadRequest, CompleteMultipartUploadRequest, CompletedMultipartUpload,
+    CompletedPart, CreateMultipartUploadRequest, S3Client, UploadPartRequest, S3,
 };
 
 use once_cell::sync::Lazy;
 
-use std::convert::From;
-use std::str::FromStr;
+use std::collections::HashMap;
 use std::sync::Mutex;
 
 use crate::s3url::*;
 use crate::s3utils::{self, WaitError};
 
+use super::OnError;
+
+const DEFAULT_MULTIPART_UPLOAD_ON_ERROR: OnError = OnError::DoNothing;
+
 struct Started {
     client: S3Client,
     buffer: Vec<u8>,
@@ -86,17 +90,64 @@ struct Settings {
     key: Option<String>,
     content_type: Option<String>,
     buffer_size: u64,
+    access_key: Option<String>,
+    secret_access_key: Option<String>,
+    metadata: Option<gst::Structure>,
+    multipart_upload_on_error: OnError,
 }
 
 impl Settings {
     fn to_uri(&self) -> String {
         format!(
             "s3://{}/{}/{}",
-            self.region.name(),
+            match self.region {
+                Region::Custom {
+                    ref name,
+                    ref endpoint,
+                } => {
+                    format!(
+                        "{}+{}",
+                        base32::encode(
+                            base32::Alphabet::RFC4648 { padding: true },
+                            name.as_bytes(),
+                        ),
+                        base32::encode(
+                            base32::Alphabet::RFC4648 { padding: true },
+                            endpoint.as_bytes(),
+                        ),
+                    )
+                }
+                _ => {
+                    String::from(self.region.name())
+                }
+            },
             self.bucket.as_ref().unwrap(),
             self.key.as_ref().unwrap()
         )
     }
+
+    fn to_metadata(&self, element: &super::S3Sink) -> Option<HashMap<String, String>> {
+        self.metadata.as_ref().map(|structure| {
+            let mut hash = HashMap::new();
+
+            for (key, value) in structure.iter() {
+                if let Ok(Ok(value_str)) = value.transform::<String>().map(|v| v.get()) {
+                    gst_log!(CAT, obj: element, "metadata '{}' -> '{}'", key, value_str);
+                    hash.insert(key.to_string(), value_str);
+                } else {
+                    gst_warning!(
+                        CAT,
+                        obj: element,
+                        "Failed to convert metadata '{}' to string ('{:?}')",
+                        key,
+                        value
+                    );
+                }
+            }
+
+            hash
+        })
+    }
 }
 
 impl Default for Settings {
@@ -107,6 +158,10 @@ impl Default for Settings {
             key: None,
             content_type: None,
             buffer_size: DEFAULT_BUFFER_SIZE,
+            access_key: None,
+            secret_access_key: None,
+            metadata: None,
+            multipart_upload_on_error: DEFAULT_MULTIPART_UPLOAD_ON_ERROR,
         }
     }
 }
@@ -117,6 +172,7 @@ pub struct S3Sink {
     settings: Mutex<Settings>,
     state: Mutex<State>,
     canceller: Mutex<Option<future::AbortHandle>>,
+    abort_multipart_canceller: Mutex<Option<future::AbortHandle>>,
 }
 
 static CAT: Lazy<gst::DebugCategory> = Lazy::new(|| {
@@ -147,10 +203,62 @@ impl S3Sink {
 
         let output =
             s3utils::wait(&self.canceller, upload_part_req_future).map_err(|err| match err {
-                WaitError::FutureError(err) => Some(gst::error_msg!(
-                    gst::ResourceError::OpenWrite,
-                    ["Failed to upload part: {}", err]
-                )),
+                WaitError::FutureError(err) => {
+                    let settings = self.settings.lock().unwrap();
+                    match settings.multipart_upload_on_error {
+                        OnError::Abort => {
+                            gst_log!(
+                                CAT,
+                                obj: element,
+                                "Aborting multipart upload request with id: {}",
+                                state.upload_id
+                            );
+                            match self.abort_multipart_upload_request(state) {
+                                Ok(()) => {
+                                    gst_log!(
+                                        CAT,
+                                        obj: element,
+                                        "Aborting multipart upload request succeeded."
+                                    );
+                                }
+                                Err(err) => gst_error!(
+                                    CAT,
+                                    obj: element,
+                                    "Aborting multipart upload failed: {}",
+                                    err.to_string()
+                                ),
+                            }
+                        }
+                        OnError::Complete => {
+                            gst_log!(
+                                CAT,
+                                obj: element,
+                                "Completing multipart upload request with id: {}",
+                                state.upload_id
+                            );
+                            match self.complete_multipart_upload_request(state) {
+                                Ok(()) => {
+                                    gst_log!(
+                                        CAT,
+                                        obj: element,
+                                        "Complete multipart upload request succeeded."
+                                    );
+                                }
+                                Err(err) => gst_error!(
+                                    CAT,
+                                    obj: element,
+                                    "Completing multipart upload failed: {}",
+                                    err.to_string()
+                                ),
+                            }
+                        }
+                        OnError::DoNothing => (),
+                    }
+                    Some(gst::error_msg!(
+                        gst::ResourceError::OpenWrite,
+                        ["Failed to upload part: {}", err]
+                    ))
+                }
                 WaitError::Cancelled => None,
             })?;
 
@@ -219,10 +327,74 @@ impl S3Sink {
             bucket: url.bucket.clone(),
             key: url.object.clone(),
             content_type: settings.content_type.clone(),
+            metadata: settings.to_metadata(&self.instance()),
             ..Default::default()
         }
     }
 
+    fn create_abort_multipart_upload_request(
+        &self,
+        url: &GstS3Url,
+        started_state: &Started,
+    ) -> AbortMultipartUploadRequest {
+        AbortMultipartUploadRequest {
+            bucket: url.bucket.clone(),
+            expected_bucket_owner: None,
+            key: url.object.clone(),
+            request_payer: None,
+            upload_id: started_state.upload_id.to_owned(),
+        }
+    }
+
+    fn abort_multipart_upload_request(
+        &self,
+        started_state: &Started,
+    ) -> Result<(), gst::ErrorMessage> {
+        let s3url = match *self.url.lock().unwrap() {
+            Some(ref url) => url.clone(),
+            None => unreachable!("Element should be started"),
+        };
+        let abort_req = self.create_abort_multipart_upload_request(&s3url, started_state);
+        let abort_req_future = started_state.client.abort_multipart_upload(abort_req);
+
+        s3utils::wait(&self.abort_multipart_canceller, abort_req_future)
+            .map(|_| ())
+            .map_err(|err| match err {
+                WaitError::FutureError(err) => {
+                    gst::error_msg!(
+                        gst::ResourceError::Write,
+                        ["Failed to abort multipart upload: {}.", err.to_string()]
+                    )
+                }
+                WaitError::Cancelled => {
+                    gst::error_msg!(
+                        gst::ResourceError::Write,
+                        ["Abort multipart upload request interrupted."]
+                    )
+                }
+            })
+    }
+
+    fn complete_multipart_upload_request(
+        &self,
+        started_state: &mut Started,
+    ) -> Result<(), gst::ErrorMessage> {
+        let complete_req = self.create_complete_multipart_upload_request(started_state);
+        let complete_req_future = started_state.client.complete_multipart_upload(complete_req);
+
+        s3utils::wait(&self.canceller, complete_req_future)
+            .map(|_| ())
+            .map_err(|err| match err {
+                WaitError::FutureError(err) => gst::error_msg!(
+                    gst::ResourceError::Write,
+                    ["Failed to complete multipart upload: {}.", err.to_string()]
+                ),
+                WaitError::Cancelled => {
+                    gst::error_msg!(gst::LibraryError::Failed, ["Interrupted during stop"])
+                }
+            })
+    }
+
     fn finalize_upload(&self, element: &super::S3Sink) -> Result<(), gst::ErrorMessage> {
         if self.flush_current_buffer(element).is_err() {
             return Err(gst::error_msg!(
@@ -239,20 +411,7 @@ impl S3Sink {
             }
         };
 
-        let complete_req = self.create_complete_multipart_upload_request(started_state);
-        let complete_req_future = started_state.client.complete_multipart_upload(complete_req);
-
-        s3utils::wait(&self.canceller, complete_req_future)
-            .map(|_| ())
-            .map_err(|err| match err {
-                WaitError::FutureError(err) => gst::error_msg!(
-                    gst::ResourceError::Write,
-                    ["Failed to complete multipart upload: {}.", err.to_string()]
-                ),
-                WaitError::Cancelled => {
-                    gst::error_msg!(gst::LibraryError::Failed, ["Interrupted during stop"])
-                }
-            })
+        self.complete_multipart_upload_request(started_state)
     }
 
     fn start(&self) -> Result<(), gst::ErrorMessage> {
@@ -273,7 +432,21 @@ impl S3Sink {
             }
         };
 
-        let client = S3Client::new(s3url.region.clone());
+        let client = match (
+            settings.access_key.as_ref(),
+            settings.secret_access_key.as_ref(),
+        ) {
+            (Some(access_key), Some(secret_access_key)) => {
+                let creds =
+                    StaticProvider::new_minimal(access_key.clone(), secret_access_key.clone());
+                S3Client::new_with(
+                    HttpClient::new().expect("failed to create request dispatcher"),
+                    creds,
+                    s3url.region.clone(),
+                )
+            }
+            _ => S3Client::new(s3url.region.clone()),
+        };
 
         let create_multipart_req = self.create_create_multipart_upload_request(&s3url, &settings);
         let create_multipart_req_future = client.create_multipart_upload(create_multipart_req);
@@ -342,6 +515,11 @@ impl S3Sink {
 
     fn cancel(&self) {
         let mut canceller = self.canceller.lock().unwrap();
+        let mut abort_canceller = self.abort_multipart_canceller.lock().unwrap();
+
+        if let Some(c) = abort_canceller.take() {
+            c.abort()
+        };
 
         if let Some(c) = canceller.take() {
             c.abort()
@@ -397,28 +575,28 @@ impl ObjectImpl for S3Sink {
     fn properties() -> &'static [glib::ParamSpec] {
         static PROPERTIES: Lazy<Vec<glib::ParamSpec>> = Lazy::new(|| {
             vec![
-                glib::ParamSpec::new_string(
+                glib::ParamSpecString::new(
                     "bucket",
                     "S3 Bucket",
                     "The bucket of the file to write",
                     None,
                     glib::ParamFlags::READWRITE | gst::PARAM_FLAG_MUTABLE_READY,
                 ),
-                glib::ParamSpec::new_string(
+                glib::ParamSpecString::new(
                     "key",
                     "S3 Key",
                     "The key of the file to write",
                     None,
                     glib::ParamFlags::READWRITE | gst::PARAM_FLAG_MUTABLE_READY,
                 ),
-                glib::ParamSpec::new_string(
+                glib::ParamSpecString::new(
                     "region",
                     "AWS Region",
                     "An AWS region (e.g. eu-west-2).",
                     None,
                     glib::ParamFlags::READWRITE | gst::PARAM_FLAG_MUTABLE_READY,
                 ),
-                glib::ParamSpec::new_uint64(
+                glib::ParamSpecUInt64::new(
                     "part-size",
                     "Part size",
                     "A size (in bytes) of an individual part used for multipart upload.",
@@ -427,13 +605,42 @@ impl ObjectImpl for S3Sink {
                     DEFAULT_BUFFER_SIZE,
                     glib::ParamFlags::READWRITE | gst::PARAM_FLAG_MUTABLE_READY,
                 ),
-                glib::ParamSpec::new_string(
+                glib::ParamSpecString::new(
                     "uri",
                     "URI",
                     "The S3 object URI",
                     None,
                     glib::ParamFlags::READWRITE | gst::PARAM_FLAG_MUTABLE_READY,
                 ),
+                glib::ParamSpecString::new(
+                    "access-key",
+                    "Access Key",
+                    "AWS Access Key",
+                    None,
+                    glib::ParamFlags::READWRITE | gst::PARAM_FLAG_MUTABLE_READY,
+                ),
+                glib::ParamSpecString::new(
+                    "secret-access-key",
+                    "Secret Access Key",
+                    "AWS Secret Access Key",
+                    None,
+                    glib::ParamFlags::READWRITE | gst::PARAM_FLAG_MUTABLE_READY,
+                ),
+                glib::ParamSpecBoxed::new(
+                    "metadata",
+                    "Metadata",
+                    "A map of metadata to store with the object in S3; field values need to be convertible to strings.",
+                    gst::Structure::static_type(),
+                    glib::ParamFlags::READWRITE | gst::PARAM_FLAG_MUTABLE_READY,
+                ),
+                glib::ParamSpecEnum::new(
+                    "on-error",
+                    "Whether to upload or complete the multipart upload on error",
+                    "Do nothing, abort or complete a multipart upload request on error",
+                    OnError::static_type(),
+                    DEFAULT_MULTIPART_UPLOAD_ON_ERROR as i32,
+                    glib::ParamFlags::READWRITE | gst::PARAM_FLAG_MUTABLE_READY,
+                ),
             ]
         });
 
@@ -449,6 +656,14 @@ impl ObjectImpl for S3Sink {
     ) {
         let mut settings = self.settings.lock().unwrap();
 
+        gst_debug!(
+            CAT,
+            obj: obj,
+            "Setting property '{}' to '{:?}'",
+            pspec.name(),
+            value
+        );
+
         match pspec.name() {
             "bucket" => {
                 settings.bucket = value
@@ -467,9 +682,18 @@ impl ObjectImpl for S3Sink {
                 }
             }
             "region" => {
-                settings.region =
-                    Region::from_str(&value.get::<String>().expect("type checked upstream"))
-                        .unwrap();
+                let region = value.get::<String>().expect("type checked upstream");
+                settings.region = region
+                    .parse::<Region>()
+                    .or_else(|_| {
+                        let (name, endpoint) = region.split_once('+').ok_or(())?;
+                        Ok(Region::Custom {
+                            name: name.into(),
+                            endpoint: endpoint.into(),
+                        })
+                    })
+                    .unwrap_or_else(|_: ()| panic!("Invalid region '{}'", region));
+
                 if settings.key.is_some() && settings.bucket.is_some() {
                     let _ = self.set_uri(obj, Some(&settings.to_uri()));
                 }
@@ -480,6 +704,19 @@ impl ObjectImpl for S3Sink {
             "uri" => {
                 let _ = self.set_uri(obj, value.get().expect("type checked upstream"));
             }
+            "access-key" => {
+                settings.access_key = value.get().expect("type checked upstream");
+            }
+            "secret-access-key" => {
+                settings.secret_access_key = value.get().expect("type checked upstream");
+            }
+            "metadata" => {
+                settings.metadata = value.get().expect("type checked upstream");
+            }
+            "on-error" => {
+                settings.multipart_upload_on_error =
+                    value.get::<OnError>().expect("type checked upstream");
+            }
             _ => unimplemented!(),
         }
     }
@@ -500,11 +737,17 @@ impl ObjectImpl for S3Sink {
 
                 url.to_value()
             }
+            "access-key" => settings.access_key.to_value(),
+            "secret-access-key" => settings.secret_access_key.to_value(),
+            "metadata" => settings.metadata.to_value(),
+            "on-error" => settings.multipart_upload_on_error.to_value(),
             _ => unimplemented!(),
         }
     }
 }
 
+impl GstObjectImpl for S3Sink {}
+
 impl ElementImpl for S3Sink {
     fn metadata() -> Option<&'static gst::subclass::ElementMetadata> {
         static ELEMENT_METADATA: Lazy<gst::subclass::ElementMetadata> = Lazy::new(|| {
